#include <windows.h>
#include <stdio.h>
#include <commctrl.h>
#include "resource.h"
#include "DGA.h"

/* Function Pointers */
typedef void (CALLBACK* PFNSETHHOOK)(HHOOK);
typedef LRESULT (CALLBACK* PFNKEYBOARDPROC)(int, WPARAM, LPARAM);
typedef void (CALLBACK* PFNNVINIT)(HWND, HINSTANCE, LPSTR, int);

/* Functions Declarations */
void CheckCmdLine(void);
LRESULT CALLBACK MainWndProc(HWND hWnd, UINT iMsg, WPARAM wParam, LPARAM lParam);
BOOL InitSettings(void);
BOOL WriteSettings(unsigned int iGamma);
void SetGamma(HWND hWnd);
void DisplayWinError(char *szTitle);
LRESULT CALLBACK AboutDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

/* Application-defined Messages */
#define WM_SYSTRAY (WM_USER + 0)

/* Constants */
static int iCtlMargin = 8;
static int iCtlDescGap = 12;
static int iCtlSliderW = 250;
static int iCtlSliderH = 34;
static char szAppName[] = APPNAME;
static char szAppDesc[] = APPDESC;
static char szLabel1[] = "Tool is now active.";
static char szLabel2[] = "Use the following keys to change gamma level:";
static char szKey1[] = "Ctrl Shift Numpad +";
static char szKeyDesc1[] = "Increase by 5%";
static char szKey2[] = "Ctrl Shift Numpad -";
static char szKeyDesc2[] = "Decrease by 5%";
static char szKey3[] = "Ctrl Shift Numpad /";
static char szKeyDesc3[] = "Restore default setting";
static char szDGADLL[] = "DGA.DLL";
static char szDetDLL[] = "NVCPL.DLL";
static char szDetDLLFunc[] = "NvStartup";
static char szDetDLLParam[] = "";
static char szRegistryKeyPart1[] = "Software\\NVIDIA Corporation\\Global\\NVTweak\\Devices";
static char szRegistryKeyPart2[] = "Color\\Schemes\\Startup";

/* Variables */
unsigned int iGammaOriginal;
unsigned int iGamma;
unsigned long lSettingsLen;
HINSTANCE hInstanceMain;
HMODULE hDetDLL;
HMODULE hDGADLL;
BOOL bStartupMinimize;
BOOL bTrayActive;
char szRegistryKey[MAX_PATH];
BYTE *a_xSettings;
PFNSETHHOOK SetHHOOK;
PFNKEYBOARDPROC KeyboardProc;
PFNNVINIT NVInit;

/* WinMain - The entry point, and main loop */
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int iCmdShow)
{
	int iReturnVal;
	HWND hWnd;
	INITCOMMONCONTROLSEX iccInit;
	WNDCLASSEX wcexMain;
	MSG msgMain;

	/* Initialize */
	hInstanceMain = hInstance;
	iReturnVal = -1;
	
	/* Check command line */
	CheckCmdLine();
	if (bStartupMinimize)
		iCmdShow = SW_SHOWMINIMIZED;
	
	/* Init the common controls library */
	iccInit.dwSize = sizeof(INITCOMMONCONTROLSEX);
	iccInit.dwICC = ICC_BAR_CLASSES;
	if (!InitCommonControlsEx(&iccInit))
	{
		DisplayWinError("Unable to init Common Controls");
		goto WinMain_Exit;
	}

	/* Load DGA library */
	hDGADLL = LoadLibrary(szDGADLL);
	if (!hDGADLL)
	{
		DisplayWinError("Unable to load DGA library");
		goto WinMain_Exit;
	}
	SetHHOOK = (PFNSETHHOOK)GetProcAddress(hDGADLL, "SetHHOOK");
	if (!SetHHOOK)
	{
		DisplayWinError("DGA library function not found: SetHHOOK");
		goto WinMain_Exit;
	}
	KeyboardProc = (PFNKEYBOARDPROC)GetProcAddress(hDGADLL, "KeyboardProc");
	if (!KeyboardProc)
	{
		DisplayWinError("DGA library function not found: KeyboardProc");
		goto WinMain_Exit;
	}

	/* Load Detonator tweak library */
	hDetDLL = LoadLibrary(szDetDLL);
	if (!hDetDLL)
	{
		DisplayWinError("Unable to load Detonator library");
		goto WinMain_Exit;
	}
	NVInit = (PFNNVINIT)GetProcAddress(hDetDLL, szDetDLLFunc);
	if (!NVInit)
	{
		DisplayWinError(szDetDLLFunc);
		goto WinMain_Exit;
	}

	/* Fetch original gamma settings from registry */
	if (!InitSettings())
		goto WinMain_Exit;

	/* Register class */
	wcexMain.cbSize = sizeof(WNDCLASSEX); 
	wcexMain.style = CS_HREDRAW | CS_VREDRAW;
	wcexMain.lpfnWndProc = MainWndProc;
	wcexMain.cbClsExtra = 0;
	wcexMain.cbWndExtra = 0;
	wcexMain.hInstance = hInstance;
	wcexMain.hIcon = LoadIcon(hInstance, (LPCTSTR)IDI_DGA);
	wcexMain.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcexMain.hbrBackground = (HBRUSH)(COLOR_BTNFACE+1);
	wcexMain.lpszMenuName = NULL;
	wcexMain.lpszClassName = szAppName;
	wcexMain.hIconSm = LoadIcon(hInstance, (LPCTSTR)IDI_DGA);
	if (!RegisterClassEx(&wcexMain))
	{
		DisplayWinError("Unable to register main window");
		goto WinMain_Exit;
	}

	/* Create window */
	hWnd = CreateWindow(szAppName, szAppDesc, WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX, CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);
	if (!hWnd)
	{
		DisplayWinError("Unable to create main window");
		goto WinMain_Exit;
	}

	/* Restore original gamma setting (just in case) */
	SetGamma(hWnd);

	/* Show window */
	ShowWindow(hWnd, iCmdShow);
	UpdateWindow(hWnd);

	/* Message loop */
	while (GetMessage(&msgMain, NULL, 0, 0))
	{
		TranslateMessage(&msgMain);
		DispatchMessage(&msgMain);
	}
	iReturnVal = msgMain.wParam;

	/* Restore gamma setting before exiting */
	WriteSettings(iGammaOriginal);
	SetGamma(hWnd);

WinMain_Exit:
	/* Free Detonator tweak library */
	if (hDetDLL)
		FreeLibrary(hDetDLL);
	
	/* Free DGA library */
	if (hDGADLL)
		FreeLibrary(hDGADLL);

	return (iReturnVal);
}

/* Check the command line */
void CheckCmdLine(void)
{
	long lOptionLen;
	char *szCmdLine;
	char *szPosE;
	char szOption[255];

	/* Initialise */
	bStartupMinimize = false;

	/* Address start of the command line */
	szCmdLine = GetCommandLine();
	if (!szCmdLine)
		goto CheckCmdLine_Exit;

	/* Bump past application filename */
	if (*szCmdLine == 34)	// Starts with a quote (may have spaces in path/file)
	{
		szCmdLine++;
		szCmdLine = strchr(szCmdLine, 34);	// Find last quote (")
	}
	else					// Else just look for first space
	{
		szCmdLine = strchr(szCmdLine, 32);	// Find first space
	}
	if (!szCmdLine)
		goto CheckCmdLine_Exit;
	szCmdLine++;

	/* Look for any command line options */
	while (strlen(szCmdLine))
	{
		/* Bump past any spaces */
		while ((strlen(szCmdLine)) && (*szCmdLine == 32))
			szCmdLine++;
		if (strlen(szCmdLine) < 1)
			break;

		/* Determine next option */
		szPosE = strchr(szCmdLine, 32);
		if (szPosE)
			lOptionLen = szPosE - szCmdLine;
		else
			lOptionLen = strlen(szCmdLine);
		strncpy(szOption, szCmdLine, lOptionLen);
		szOption[lOptionLen] = 0;
		szCmdLine += lOptionLen;

		/* Process option */
		if (stricmp(szOption, "/STARTUP") == 0)
			bStartupMinimize = true;
	}

CheckCmdLine_Exit:
	return;
}

/* Message Handler */
LRESULT CALLBACK MainWndProc(HWND hWnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	int iCtlXPos, iCtlYPos;
	unsigned int iScrollReq;
	unsigned int iGammaNew;
	static HWND hWndLabel1, hWndLabel2, hWndKey1, hWndKeyDesc1, hWndKey2, hWndKeyDesc2, hWndKey3, hWndKeyDesc3, hWndSlider1, hWndAbout;
	HDC hDCCtl;
	HFONT hFontCtl;
	static HHOOK hKeyHook;
	BOOL bForceUpdate;
	SIZE sizFontCtl;
	RECT recWin, recCli;
	static NOTIFYICONDATA nidDGA;
	PCOPYDATASTRUCT pcdDGA;
	
	switch (iMsg)
	{
		/* Create window */
		case WM_CREATE:
			/* Get the font to be used */
 			hFontCtl = (HFONT)GetStockObject(ANSI_VAR_FONT);

			/* Add the static controls */
			iCtlYPos = iCtlMargin;
			hWndLabel1 = CreateWindow("STATIC", szLabel1, WS_CHILD | WS_VISIBLE, 1, 1, 1, 1, hWnd, NULL, hInstanceMain, NULL);
			SendMessage(hWndLabel1, WM_SETFONT, (WPARAM)hFontCtl, false);
			hDCCtl = GetDC(hWndLabel1);
			SelectObject(hDCCtl, hFontCtl);
			GetTextExtentPoint32(hDCCtl, szLabel1, strlen(szLabel1), &sizFontCtl);
			ReleaseDC(hWndLabel1, hDCCtl);
			MoveWindow(hWndLabel1, iCtlMargin, iCtlYPos, sizFontCtl.cx, sizFontCtl.cy, false);
			iCtlYPos += sizFontCtl.cy + iCtlMargin;

			hWndLabel2 = CreateWindow("STATIC", szLabel2, WS_CHILD | WS_VISIBLE, 1, 1, 1, 1, hWnd, NULL, hInstanceMain, NULL);
			SendMessage(hWndLabel2, WM_SETFONT, (WPARAM)hFontCtl, false);
			hDCCtl = GetDC(hWndLabel2);
			SelectObject(hDCCtl, hFontCtl);
			GetTextExtentPoint32(hDCCtl, szLabel2, strlen(szLabel2), &sizFontCtl);
			ReleaseDC(hWndLabel2, hDCCtl);
			MoveWindow(hWndLabel2, iCtlMargin, iCtlYPos, sizFontCtl.cx, sizFontCtl.cy, false);
			iCtlYPos += sizFontCtl.cy + (iCtlMargin / 2);

			hWndKey1 = CreateWindow("STATIC", szKey1, WS_CHILD | WS_VISIBLE, 1, 1, 1, 1, hWnd, NULL, hInstanceMain, NULL);
			SendMessage(hWndKey1, WM_SETFONT, (WPARAM)hFontCtl, false);
			hDCCtl = GetDC(hWndKey1);
			SelectObject(hDCCtl, hFontCtl);
			GetTextExtentPoint32(hDCCtl, szKey1, strlen(szKey1), &sizFontCtl);
			ReleaseDC(hWndKey1, hDCCtl);
			MoveWindow(hWndKey1, (iCtlMargin * 2), iCtlYPos, sizFontCtl.cx, sizFontCtl.cy, false);
			iCtlXPos = (iCtlMargin * 2) + sizFontCtl.cx + iCtlDescGap;
			hWndKeyDesc1 = CreateWindow("STATIC", szKeyDesc1, WS_CHILD | WS_VISIBLE, 1, 1, 1, 1, hWnd, NULL, hInstanceMain, NULL);
			SendMessage(hWndKeyDesc1, WM_SETFONT, (WPARAM)hFontCtl, false);
			hDCCtl = GetDC(hWndKeyDesc1);
			SelectObject(hDCCtl, hFontCtl);
			GetTextExtentPoint32(hDCCtl, szKeyDesc1, strlen(szKeyDesc1), &sizFontCtl);
			ReleaseDC(hWndKeyDesc1, hDCCtl);
			MoveWindow(hWndKeyDesc1, iCtlXPos, iCtlYPos, sizFontCtl.cx, sizFontCtl.cy, false);
			iCtlYPos += sizFontCtl.cy;

			hWndKey2 = CreateWindow("STATIC", szKey2, WS_CHILD | WS_VISIBLE, 1, 1, 1, 1, hWnd, NULL, hInstanceMain, NULL);
			SendMessage(hWndKey2, WM_SETFONT, (WPARAM)hFontCtl, false);
			hDCCtl = GetDC(hWndKey2);
			SelectObject(hDCCtl, hFontCtl);
			GetTextExtentPoint32(hDCCtl, szKey2, strlen(szKey2), &sizFontCtl);
			ReleaseDC(hWndKey2, hDCCtl);
			MoveWindow(hWndKey2, (iCtlMargin * 2), iCtlYPos, sizFontCtl.cx, sizFontCtl.cy, false);
			hWndKeyDesc2 = CreateWindow("STATIC", szKeyDesc2, WS_CHILD | WS_VISIBLE, 1, 1, 1, 1, hWnd, NULL, hInstanceMain, NULL);
			SendMessage(hWndKeyDesc2, WM_SETFONT, (WPARAM)hFontCtl, false);
			hDCCtl = GetDC(hWndKeyDesc2);
			SelectObject(hDCCtl, hFontCtl);
			GetTextExtentPoint32(hDCCtl, szKeyDesc2, strlen(szKeyDesc2), &sizFontCtl);
			ReleaseDC(hWndKeyDesc2, hDCCtl);
			MoveWindow(hWndKeyDesc2, iCtlXPos, iCtlYPos, sizFontCtl.cx, sizFontCtl.cy, false);
			iCtlYPos += sizFontCtl.cy;

			hWndKey3 = CreateWindow("STATIC", szKey3, WS_CHILD | WS_VISIBLE, 1, 1, 1, 1, hWnd, NULL, hInstanceMain, NULL);
			SendMessage(hWndKey3, WM_SETFONT, (WPARAM)hFontCtl, false);
			hDCCtl = GetDC(hWndKey3);
			SelectObject(hDCCtl, hFontCtl);
			GetTextExtentPoint32(hDCCtl, szKey3, strlen(szKey3), &sizFontCtl);
			ReleaseDC(hWndKey3, hDCCtl);
			MoveWindow(hWndKey3, (iCtlMargin * 2), iCtlYPos, sizFontCtl.cx, sizFontCtl.cy, false);
			hWndKeyDesc3 = CreateWindow("STATIC", szKeyDesc3, WS_CHILD | WS_VISIBLE, 1, 1, 1, 1, hWnd, NULL, hInstanceMain, NULL);
			SendMessage(hWndKeyDesc3, WM_SETFONT, (WPARAM)hFontCtl, false);
			hDCCtl = GetDC(hWndKeyDesc3);
			SelectObject(hDCCtl, hFontCtl);
			GetTextExtentPoint32(hDCCtl, szKeyDesc3, strlen(szKeyDesc3), &sizFontCtl);
			ReleaseDC(hWndKeyDesc3, hDCCtl);
			MoveWindow(hWndKeyDesc3, iCtlXPos, iCtlYPos, sizFontCtl.cx, sizFontCtl.cy, false);
			iCtlYPos += sizFontCtl.cy + iCtlMargin;

			/* Add the slider control */
			hWndSlider1 = CreateWindow(TRACKBAR_CLASS, "", WS_CHILD | WS_VISIBLE | TBS_AUTOTICKS | TBS_HORZ | TBS_TOOLTIPS | TBS_TOP, iCtlMargin, iCtlYPos, iCtlSliderW, iCtlSliderH, hWnd, NULL, hInstanceMain, NULL);
			SendMessage(hWndSlider1, WM_SETFONT, (WPARAM)hFontCtl, false);
			SendMessage(hWndSlider1, TBM_SETRANGE, false, MAKELONG(0, 100));
			SendMessage(hWndSlider1, TBM_SETTICFREQ, 5, 0);
			SendMessage(hWndSlider1, TBM_SETLINESIZE, 0, 1);
			SendMessage(hWndSlider1, TBM_SETPAGESIZE, 0, 5);
			SendMessage(hWndSlider1, TBM_SETPOS, true, (LPARAM)iGammaOriginal);
			iCtlYPos += iCtlSliderH + iCtlMargin;

			/* Adjust the window dimensions to match controls */
			GetWindowRect(hWnd, &recWin);
			GetClientRect(hWnd, &recCli);
			MoveWindow(hWnd, recWin.left, recWin.top, iCtlSliderW + (2 * iCtlMargin) + ((recWin.right - recWin.left) - (recCli.right - recCli.left)), iCtlYPos + ((recWin.bottom - recWin.top) - (recCli.bottom - recCli.top)), false);

			/* Add the About button control */
			GetClientRect(hWnd, &recCli);
			hWndAbout = CreateWindowEx(WS_EX_STATICEDGE, "BUTTON", "About", WS_CHILD | WS_VISIBLE, (recCli.right - recCli.left) - iCtlMargin/2 - 40, iCtlMargin/2, 40, 20, hWnd, NULL, hInstanceMain, NULL);
			SendMessage(hWndAbout, WM_SETFONT, (WPARAM)hFontCtl, false);

			/* Init gamma setting */
			iGamma = iGammaOriginal;

			/* Build 'Notify Icon Data' for system tray operations */
			nidDGA.cbSize = sizeof(NOTIFYICONDATA);
			nidDGA.hWnd = hWnd;
			nidDGA.uID = 0;
			nidDGA.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
			nidDGA.uCallbackMessage = WM_SYSTRAY;
			nidDGA.hIcon = LoadIcon(hInstanceMain, (LPCTSTR)IDI_DGA);
			sprintf(nidDGA.szTip, "%s [%d%%]", szAppDesc, iGamma);

			/* Start the keyboard hook */
			hKeyHook = SetWindowsHookEx(WH_KEYBOARD, KeyboardProc, hDGADLL, 0);
			if (!hKeyHook)
			{
				DisplayWinError("Failed attempt to place keyboard hook");
			}
			
			/* Pass the hook handle to the DGA library */
			SetHHOOK(hKeyHook);
			
			return 0;

		/* Window got focus */
		case WM_SETFOCUS:
			/* Give focus to the slider control */
			SetFocus(hWndSlider1);
			
			return 0;

		/* Window size change */
		case WM_SIZE:
			switch (wParam)
			{
				/* Restored to normal size */
				case SIZE_RESTORED:
					if (bStartupMinimize)
						break;
					Shell_NotifyIcon(NIM_DELETE, &nidDGA);
					ShowWindow(hWnd, SW_SHOW);
					bTrayActive = false;
					return 0;
				/* Minimized */
				case SIZE_MINIMIZED:
					Shell_NotifyIcon(NIM_ADD, &nidDGA);
					ShowWindow(hWnd, SW_HIDE);
					bTrayActive = true;
					if (bStartupMinimize)
						bStartupMinimize = false;
					return 0;
			}

		/* Events while minimized in the system tray */
		case WM_SYSTRAY:
			switch (lParam)
			{
				case WM_LBUTTONUP:
					ShowWindow(hWnd, SW_SHOW);
					ShowWindow(hWnd, SW_RESTORE);
					break;
			}
			return 0;

		/* Horizontal scroll */
		case WM_HSCROLL:
			/* Ensure it comes from the gamma slider control */
			if ((HWND)lParam != hWndSlider1)
				break;

			/* Only process when user has completed scroll */
			iScrollReq = LOWORD(wParam);
			if (iScrollReq != TB_ENDTRACK)
				break;

			/* Get new gamma setting, and exit if no change */
			iGammaNew = SendMessage(hWndSlider1, TBM_GETPOS, 0, 0);
			if (iGammaNew == iGamma)
				break;
			iGamma = iGammaNew;
			sprintf(nidDGA.szTip, "%s [%d%%]", szAppDesc, iGamma);
			if (bTrayActive)
				Shell_NotifyIcon(NIM_MODIFY, &nidDGA);

			/* Update gamma setting */
			WriteSettings(iGamma);
			SetGamma(hWnd);

			return 0;

		/* Data passed from DLL */
		case WM_COPYDATA:
			/* Determine new gamma setting */
			pcdDGA = (PCOPYDATASTRUCT)lParam;
			switch (pcdDGA->dwData)
			{
				case VK_DIVIDE:
					iGammaNew = iGammaOriginal;
					bForceUpdate = true;
					break;
				case VK_SUBTRACT:
					if (iGamma < 5)
						iGammaNew = 0;
					else
						iGammaNew = iGamma - 5;
					bForceUpdate = false;
					break;
				case VK_ADD:
					if (iGamma > 95)
						iGammaNew = 100;
					else
						iGammaNew = iGamma + 5;
					break;
					bForceUpdate = false;
				default:	// Unknown data...
					return false;	// ...exit
			}

			/* Ensure gamma setting has changed before continuing */
			if ((iGammaNew == iGamma) && !bForceUpdate)
				break;
			iGamma = iGammaNew;
			
			/* Update slider and systray description */
			SendMessage(hWndSlider1, TBM_SETPOS, true, (LPARAM)iGamma);
			sprintf(nidDGA.szTip, "%s [%d%%]", szAppDesc, iGamma);
			if (bTrayActive)
				Shell_NotifyIcon(NIM_MODIFY, &nidDGA);

			/* Update gamma setting */
			WriteSettings(iGamma);
			SetGamma(hWnd);

			return true;

		/* Commands */
		case WM_COMMAND:
			/* About button */
			if ((HWND)lParam == hWndAbout)
			{
				DialogBox(hInstanceMain, MAKEINTRESOURCE(IDD_ABOUT), hWnd, (DLGPROC)AboutDialogProc);
				return 0;
			}
			break;

		/* Destroy window */
		case WM_DESTROY:
			/* Start the keyboard hook */
			if (!UnhookWindowsHookEx(hKeyHook))
			{
				DisplayWinError("Failed attempt to remove keyboard hook");
			}

			/* Post quit message to WinMain message loop */
			PostQuitMessage(0);

			return 0;
	}

	return (DefWindowProc(hWnd, iMsg, wParam, lParam));
}

/* Initialise gamma settings from registry */
BOOL InitSettings(void)
{
	unsigned int iGammaR, iGammaG, iGammaB;
	unsigned long lKeyLen;
	HKEY hRegKey;
	static char szErrorTitle[] = "Unable to fetch gamma settings from registry";
	char szKey[MAX_PATH];
	
	/* Enumerate registry key to determine 'Device' name */
	if (RegOpenKeyEx(HKEY_CURRENT_USER, szRegistryKeyPart1, 0, KEY_ALL_ACCESS, &hRegKey) != ERROR_SUCCESS)
	{
		DisplayWinError(szErrorTitle);
		return false;
	}
	lKeyLen = MAX_PATH;
	if (RegEnumKeyEx(hRegKey, 0, szKey, &lKeyLen, NULL, NULL, NULL, NULL) != ERROR_SUCCESS)
	{
		DisplayWinError(szErrorTitle);
		return false;
	}
	if (RegCloseKey(hRegKey) != ERROR_SUCCESS)
	{
		DisplayWinError(szErrorTitle);
		return false;
	}

	/* Build registry path, incorporating the device name */
	sprintf(szRegistryKey, "%s\\%s\\%s", szRegistryKeyPart1, szKey, szRegistryKeyPart2);
	
	/* Enumerate registry key to determine 'Startup' name */
	if (RegOpenKeyEx(HKEY_CURRENT_USER, szRegistryKey, 0, KEY_ALL_ACCESS, &hRegKey) != ERROR_SUCCESS)
	{
		DisplayWinError(szErrorTitle);
		return false;
	}
	lKeyLen = MAX_PATH;
	if (RegEnumKeyEx(hRegKey, 0, szKey, &lKeyLen, NULL, NULL, NULL, NULL) != ERROR_SUCCESS)
	{
		DisplayWinError(szErrorTitle);
		return false;
	}
	if (RegCloseKey(hRegKey) != ERROR_SUCCESS)
	{
		DisplayWinError(szErrorTitle);
		return false;
	}

	/* Append 'Startup' name to registry path */
	strcat(szRegistryKey, "\\");
	strcat(szRegistryKey, szKey);
	
	/* Open registry key */
	if (RegOpenKeyEx(HKEY_CURRENT_USER, szRegistryKey, 0, KEY_ALL_ACCESS, &hRegKey) != ERROR_SUCCESS)
	{
		DisplayWinError(szErrorTitle);
		return false;
	}

	/* Determine size of color settings data */
	if (RegQueryValueEx(hRegKey, "CurrentValue", NULL, NULL, NULL, &lSettingsLen) != ERROR_SUCCESS)
	{
		DisplayWinError(szErrorTitle);
		return false;
	}

	/* Allocate space for data and fetch it */
	a_xSettings = (BYTE *)malloc(lSettingsLen);
	if (RegQueryValueEx(hRegKey, "CurrentValue", NULL, NULL, a_xSettings, &lSettingsLen) != ERROR_SUCCESS)
	{
		DisplayWinError(szErrorTitle);
		free(a_xSettings);
		return false;
	}

	/* Check red/green/blue gamma are identical - we need this */
	iGammaR = *(unsigned int *)(a_xSettings + 0x18);
	iGammaG = *(unsigned int *)(a_xSettings + 0x1C);
	iGammaB = *(unsigned int *)(a_xSettings + 0x20);
	if ((iGammaR != iGammaG) || (iGammaR != iGammaB))
	{
		MessageBox(NULL, "Gamma settings differ for Red/Green/Blue!\n\nThis tool only works when all colors use the same gamma setting.", szErrorTitle, MB_ICONERROR);
		return false;
	}

	/* Save original gamma level */
	iGammaOriginal = iGammaR;

	/* Close registry */
	if (RegCloseKey(hRegKey) != ERROR_SUCCESS)
	{
		DisplayWinError(szErrorTitle);
		return false;
	}
	
	/* Success */
	return true;
}

/* Write gamma settings to the registry */
BOOL WriteSettings(unsigned int iGamma)
{
	HKEY hRegKey;
	
	/* Create registry key */
	if (RegCreateKeyEx(HKEY_CURRENT_USER, szRegistryKey, 0, "", REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hRegKey, NULL) != ERROR_SUCCESS)
	{
		DisplayWinError("Unable to write gamma settings to registry");
		return false;
	}

	/* Update gamma level in color settings data */
	*(unsigned int *)(a_xSettings + 0x18) = iGamma;
	*(unsigned int *)(a_xSettings + 0x1C) = iGamma;
	*(unsigned int *)(a_xSettings + 0x20) = iGamma;

	/* Write color settings data */
	if (RegSetValueEx(hRegKey, "CurrentValue", 0, REG_BINARY, a_xSettings, lSettingsLen) != ERROR_SUCCESS)
	{
		DisplayWinError("Unable to write gamma settings to registry");
		return false;
	}

	/* Close registry */
	if (RegCloseKey(hRegKey) != ERROR_SUCCESS)
	{
		DisplayWinError("Unable to write gamma settings to registry");
		return false;
	}
	
	/* Success */
	return true;
}

/* Set gamma by calling Detonator init function */
void SetGamma(HWND hWnd)
{
	NVInit(hWnd, hInstanceMain, szDetDLLParam, SW_SHOWNA);
}

/* Display Windows error message */
void DisplayWinError(char *szTitle)
{
	char szErrorMsg[255];
	LPVOID lpErrMsgBuf;

	/* Init error message */
	strcpy(szErrorMsg, szTitle);

	/* Append Windows error message */
	if (FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpErrMsgBuf, 0, NULL))
	{
		strcat(szErrorMsg, "\n\nReason:\n");
		strcat(szErrorMsg, (char *)lpErrMsgBuf);
		LocalFree(lpErrMsgBuf);
	}
	
	/* Display the error message */
	MessageBox(NULL, szErrorMsg, "Error", MB_ICONERROR);
}

/* About dialog message handler */
LRESULT CALLBACK AboutDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
		/* Initialisation of dialog */
		case WM_INITDIALOG:
			return true;

		/* Process commands */
		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDC_ABOUT_OK:
					EndDialog(hDlg, 1);
					return false;
			}
			break;

		/* Process commands */
		case WM_SYSCOMMAND:
			switch (wParam)
			{
				case SC_CLOSE:
					EndDialog(hDlg, 1);
					return false;
			}
			break;
	}

	return false;
}
