#include <windows.h>
#include "..\DGA.h"

/* Functions Declarations - Exported */
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved);
void WINAPI SetHHOOK(HHOOK hHook);
LRESULT CALLBACK KeyboardProc(int iCode, WPARAM wParam, LPARAM lParam);

/* Functions Declarations - Local */
void FindDGAWindow(void);

/* Constants */
static char szAppName[] = APPNAME;
static char szAppDesc[] = APPDESC;

/* Variables */
HWND hWndDGA;
HHOOK hKeyHook;

/* DLL Main - Entry Point */
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	switch (fdwReason)
	{
		case DLL_PROCESS_ATTACH:
			FindDGAWindow();
			break;
		case DLL_PROCESS_DETACH:
			break;
	}
	return true;
}

/* Set hook handle */
void WINAPI SetHHOOK(HHOOK hHook)
{
	hKeyHook = hHook;
}

/* Keyboard hook */
LRESULT CALLBACK KeyboardProc(int iCode, WPARAM wParam, LPARAM lParam)
{
	BOOL bControlDown, bShiftDown;
	COPYDATASTRUCT cdDGA;

	/* Never process call when code is less than zero */
	if (iCode < 0)
		goto KeyboardProc_Exit;

	/* Check virtual key code is for numpad divide, subtract or add */
	if ((wParam != VK_DIVIDE) && (wParam != VK_SUBTRACT) && (wParam != VK_ADD))
		goto KeyboardProc_Exit;
	if (!(lParam & 0x80000000))		// Only process key down
		goto KeyboardProc_Exit;

	/* Ensure both the Control and Shift keys are pressed */
	bControlDown = GetKeyState(VK_CONTROL) & 0x8000;
	bShiftDown = GetKeyState(VK_SHIFT) & 0x8000;
	if (!bControlDown || !bShiftDown)
		goto KeyboardProc_Exit;

	/* If we haven't already got the DGA window handle, do so now */
	if (!hWndDGA)
		FindDGAWindow();

	/* Send appropriate message to the DGA application */
	cdDGA.dwData = wParam;
	cdDGA.cbData = 0;
	cdDGA.lpData = NULL;
	SendMessage(hWndDGA, WM_COPYDATA, 0, (LPARAM)&cdDGA);
	
KeyboardProc_Exit:
	return (CallNextHookEx(hKeyHook, iCode, wParam, lParam));
}

/* Find DGA window */
void FindDGAWindow(void)
{
	hWndDGA = FindWindow(szAppName, szAppDesc);
}
